@@ -1,12 +1,22 @@
 """Client for Omada Site requests."""
 
-from typing import List, Optional, Union
-from .omadaapiconnection import OmadaApiConnection
+from typing import Any, AsyncIterable, List, Optional, Union
 
+from .clients import (
+    OmadaClientDetails,
+    OmadaConnectedClient,
+    OmadaNetworkClient,
+    OmadaWiredClient,
+    OmadaWiredClientDetails,
+    OmadaWirelessClient,
+    OmadaWirelessClientDetails,
+)
+from .definitions import BandwidthControl, Eth802Dot1X, LinkDuplex, LinkSpeed, PoEMode
 from .devices import (
     OmadaAccessPoint,
     OmadaDevice,
     OmadaFirmwareUpdate,
+    OmadaGateway,
     OmadaListDevice,
     OmadaPortProfile,
     OmadaSwitch,
@@ -14,12 +24,10 @@ from .devices import (
     OmadaSwitchPortDetails,
     OmadaAccesPointLanPortSettings,
 )
-
 from .exceptions import (
     InvalidDevice,
 )
-
-from .definitions import BandwidthControl, Eth802Dot1X, LinkDuplex, LinkSpeed, PoEMode
+from .omadaapiconnection import OmadaApiConnection
 
 
 class SwitchPortOverrides:
@@ -78,6 +86,56 @@ class OmadaSiteClient:
         self._api = api
         self._site_id = site_id
 
+    async def block_client(self, mac_or_client: Union[str, OmadaNetworkClient]) -> None:
+        if isinstance(mac_or_client, OmadaConnectedClient):
+            mac = mac_or_client.mac
+        else:
+            mac = mac_or_client
+        await self._api.request(
+            "post", self._api.format_url(f"cmd/clients/{mac}/block", self._site_id)
+        )
+
+    async def unblock_client(self, mac_or_client: Union[str, OmadaNetworkClient]) -> None:
+        if isinstance(mac_or_client, OmadaConnectedClient):
+            mac = mac_or_client.mac
+        else:
+            mac = mac_or_client
+        await self._api.request(
+            "post", self._api.format_url(f"cmd/clients/{mac}/unblock", self._site_id)
+        )
+
+    async def get_client(self, mac_or_client: Union[str, OmadaNetworkClient]) -> OmadaClientDetails:
+        """Get the details of a client"""
+        if isinstance(mac_or_client, OmadaConnectedClient):
+            mac = mac_or_client.mac
+        else:
+            mac = mac_or_client
+
+        result = await self._api.request(
+            "get", self._api.format_url(f"clients/{mac}", self._site_id)
+        )
+
+        if result.get("wireless"):
+            return OmadaWirelessClientDetails(result)
+        else:
+            return OmadaWiredClientDetails(result)
+
+    async def get_connected_clients(self) -> AsyncIterable[OmadaConnectedClient]:
+        """Get the clients connected to the site network."""
+        async for client in self._api.iterate_pages(self._api.format_url("clients", self._site_id)):
+            if client["wireless"]:
+                yield OmadaWirelessClient(client)
+            else:
+                yield OmadaWiredClient(client)
+
+    async def get_known_clients(self) -> AsyncIterable[OmadaNetworkClient]:
+        """Get the clients connected to the site network."""
+        async for client in self._api.iterate_pages(self._api.format_url("insight/clients", self._site_id)):
+            if client["wireless"]:
+                yield OmadaWirelessClient(client)
+            else:
+                yield OmadaWiredClient(client)
+
     async def get_devices(self) -> List[OmadaListDevice]:
         """Get the list of devices on the site."""
 
@@ -142,7 +200,7 @@ class OmadaSiteClient:
     async def get_switch_ports(
         self, mac_or_device: Union[str, OmadaDevice]
     ) -> List[OmadaSwitchPortDetails]:
-        """Get a switch by Mac address or Omada device."""
+        """Get ports of a switch by Mac address or Omada device."""
 
         if isinstance(mac_or_device, OmadaDevice):
             if mac_or_device.type != "switch":
@@ -162,7 +220,8 @@ class OmadaSiteClient:
         mac_or_device: Union[str, OmadaDevice],
         index_or_port: Union[int, OmadaSwitchPort],
     ) -> OmadaSwitchPortDetails:
-        """Get a switch by Mac address or Omada device."""
+        """Get a single port of a switch by Mac address or Omada device,
+            and port number or port object of OmadaSwitch device."""
 
         if isinstance(mac_or_device, OmadaDevice):
             if mac_or_device.type != "switch":
@@ -182,6 +241,46 @@ class OmadaSiteClient:
 
         return OmadaSwitchPortDetails(result)
 
+    async def get_switch_port_overrides(
+        self,
+        mac_or_device: Union[str, OmadaDevice],
+        index_or_port: Union[int, OmadaSwitchPort],
+    ) -> SwitchPortOverrides:
+        """Return the current override settings for the port of a switch, or the current profile settings as default."""
+
+        port = await self.get_switch_port(mac_or_device, index_or_port)
+
+        # Returns the current overrides
+        if port.has_profile_override:
+            return SwitchPortOverrides(
+                    enable_poe=(port.poe_mode == PoEMode.ENABLED),
+                    dot1x_mode=port.eth_802_1x_control,
+                    duplex=port.duplex,
+                    link_speed=port.link_speed,
+                    lldp_med_enable=port.lldp_med_enabled,
+                    loopback_detect=port.loopback_detect_enabled,
+                    spanning_tree_enable=port.spanning_tree_enabled,
+                    port_isolation=port.port_isolation_enabled
+            )
+
+        # Otherwise the profile's config values are returned
+        prof = await self.get_port_profile(port.profile_id)
+        
+        # The API doesn't provide the PoE mode of the switch (couldn't even find in Omada
+        # GUI how to set the PoE mode of a switch). Thus, use True as a default value.
+        poe_mode = (prof.poe_mode != PoEMode.DISABLED)
+
+        return SwitchPortOverrides(
+                enable_poe=poe_mode,
+                dot1x_mode=prof.eth_802_1x_control,
+                duplex=LinkDuplex.AUTO,
+                link_speed=LinkSpeed.SPEED_AUTO,
+                lldp_med_enable=prof.lldp_med_enabled,
+                loopback_detect=prof.loopback_detect_enabled,
+                spanning_tree_enable=prof.spanning_tree_enabled,
+                port_isolation=prof.port_isolation_enabled
+        )
+
     async def update_access_point_port(
         self,
         mac_or_device: Union[str, OmadaDevice],
@@ -274,6 +373,15 @@ class OmadaSiteClient:
         # Read back the new port settings
         return await self.get_switch_port(mac, port)
 
+    async def get_port_profile(self, profile_id: str) -> OmadaPortProfile:
+        profiles = await self.get_port_profiles()
+
+        profile = next((p for p in profiles if p.profile_id == profile_id), None)
+
+        if not profile:
+            raise InvalidDevice(f"Port profile {profile_id} does not exist")
+        return profile        
+
     async def get_port_profiles(self) -> List[OmadaPortProfile]:
         """Lists the available switch port profiles that can be applied."""
 
@@ -317,3 +425,33 @@ class OmadaSiteClient:
         )
 
         return True
+
+    async def get_gateways(self) -> List[OmadaGateway]:
+        """Get the list of gateways (routers) on the site. (Zero or one!)"""
+
+        return [
+            await self.get_gateway(d)
+            for d in await self.get_devices()
+            if d.type == "gateway"
+        ]
+
+    async def get_gateway(self, mac_or_device: Union[str, OmadaDevice, None]) -> OmadaGateway:
+        """Get the gatway (router) for the site by Mac address or Omada device. (There can be only one!)"""
+
+        if mac_or_device is None:
+            mac_or_device = next((d for d in await self.get_devices() if d.type == "gateway"), None)
+            if mac_or_device is None:
+                raise InvalidDevice("No gateways found in site")
+
+        if isinstance(mac_or_device, OmadaDevice):
+            if mac_or_device.type != "gateway":
+                raise InvalidDevice()
+            mac = mac_or_device.mac
+        else:
+            mac = mac_or_device
+
+        result = await self._api.request(
+            "get", self._api.format_url(f"gateways/{mac}", self._site_id)
+        )
+
+        return OmadaGateway(result)
